module fourier

  implicit none
  include "fftw3.f"

  complex, dimension(:), allocatable, private  :: in1d, out1d,in1, out1, in2, out2
  complex, dimension(:,:), allocatable, private  :: in2d, out2d, temp2d
  integer, private                             :: n1, n2
  integer*8, private                           :: plan1f, plan1b,plan1f_2d, plan1b_2d, plan2f_2d, plan2b_2d
  integer*8, private                           :: plan2df_2d, plan2db_2d
  contains

  subroutine ft1d_init(n1_in)
    integer               :: n1_in
    n1 = n1_in
    allocate(in1d(n1), out1d(n1))
    in1d = 0.
    out1d = 0.
    call sfftw_plan_dft_1d(plan1f, n1, in1d, out1d, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan1b, n1, out1d, in1d, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine
  
  subroutine ft1d_destroy()
    call sfftw_destroy_plan(plan1f)
    call sfftw_destroy_plan(plan1b)
  end subroutine

  subroutine ft2d_init(n1_in, n2_in)
    integer               :: n1_in, n2_in
    n1 = n1_in
    n2 = n2_in
    allocate(in1(n1), out1(n1))
    allocate(in2(n2), out2(n2))
    in1 = 0.
    out1 = 0.
    in2 = 0.
    out2 = 0.
    call sfftw_plan_dft_1d(plan1f_2d, n1, in1, out1, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan1b_2d, n1, out1, in1, FFTW_BACKWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2f_2d, n2, in2, out2, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2b_2d, n2, out2, in2, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine

  subroutine ft2dd_init(n1_in, n2_in)
    integer               :: n1_in, n2_in
    n1 = n1_in
    n2 = n2_in
    allocate(in2d(n1,n2), out2d(n1,n2),temp2d(n1,n2))
    in2d = 0.
    out2d = 0.
    temp2d = 0.
    call sfftw_plan_dft_2d(plan2df_2d, n1, n2, in2d, out2d, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_2d(plan2db_2d, n1, n2, in2d, out2d, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine

  subroutine ft1d(adj, sign1, cdata)
    logical, intent(in)   :: adj
    complex, dimension(:) :: cdata
    real                  :: sign1
    if(adj) then
        out1d(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1)
        out1d((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2)
        call sfftw_execute_dft(plan1b, out1d, in1d)
        cdata(:) = in1d
    else
        in1d = cdata(:)
        call sfftw_execute_dft(plan1f, in1d, out1d)
        cdata(:) = out1d
        cdata(1:n1-(n1+1)/2) = out1d((n1+1)/2+1:n1)
        cdata(n1-(n1+1)/2+1:n1) = out1d(1:(n1+1)/2)
    endif
    cdata = sign1*cdata/sqrt(1.*n1)
  end subroutine

  subroutine ft1axis(adj, sign1, cdata)
    logical, intent(in)        :: adj
    integer                    :: i2
    complex, dimension(:,:) :: cdata
    real                       :: sign1
    if(adj) then
      do i2=1,n2
        out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1,i2)
        out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2,i2)
        call sfftw_execute_dft(plan1b_2d, out1, in1)
        cdata(:,i2) = in1
      enddo
    else
      do i2=1,n2
        in1 = cdata(:,i2)
        call sfftw_execute_dft(plan1f_2d, in1, out1)
        cdata(:,i2) = out1
        cdata(1:n1-(n1+1)/2,i2) = out1((n1+1)/2+1:n1)
        cdata(n1-(n1+1)/2+1:n1,i2) = out1(1:(n1+1)/2)
      enddo
    endif
    cdata = sign1*cdata/sqrt(1.*n1)
  end subroutine

  subroutine ft2axis(adj, sign2, cdata)
    logical, intent(in)        :: adj
    integer                    :: i1
    complex, dimension(:,:) :: cdata
    real                       :: sign2
    if(adj) then
      do i1=1,n1
        out2(1:(n2+1)/2) = cdata(i1,n2-(n2+1)/2+1:n2)
        out2((n2+1)/2+1:n2) = cdata(i1,1:n2-(n2+1)/2)
        call sfftw_execute_dft(plan2b_2d, out2, in2)
        cdata(i1,:) = in2
      enddo
    else
      do i1=1,n1
        in2 = cdata(i1,:)
        call sfftw_execute_dft(plan2f_2d, in2, out2)
        cdata(i1,:) = sign2*out2/sqrt(1.*n2)
        cdata(i1,1:n2-(n2+1)/2) = out2((n2+1)/2+1:n2)
        cdata(i1,n2-(n2+1)/2+1:n2) = out2(1:(n2+1)/2)
      enddo
    endif
    cdata = sign2*cdata/sqrt(1.*n2)
  end subroutine

  subroutine ft2d(adj, sign1, sign2, cdata)
    logical, intent(in)        :: adj
    integer                    :: i1
    complex, dimension(:,:) :: cdata
    real                       :: sign1
    real                       :: sign2
    if(adj) then
      temp2d(1:(n1+1)/2,:)    = cdata(n1-(n1+1)/2+1:n1,:)
      temp2d((n1+1)/2+1:n1,:) = cdata(1:n1-(n1+1)/2,:)
      out2d(:,1:(n2+1)/2)     = temp2d(:,n2-(n2+1)/2+1:n2)
      out2d(:,(n2+1)/2+1:n2)  = temp2d(:,1:n2-(n2+1)/2)
      call sfftw_execute_dft(plan2db_2d, out2d, in2d)
      cdata(:,:) = in2d
    else
      in2d = cdata(:,:)
      call sfftw_execute_dft(plan2df_2d, in2d, out2d)
      cdata(:,:) = out2d
      temp2d(1:n1-(n1+1)/2,:)    = out2d((n1+1)/2+1:n1,:)
      temp2d(n1-(n1+1)/2+1:n1,:) = out2d(1:(n1+1)/2,:)
      cdata(:,1:n2-(n2+1)/2)     = temp2d(:,(n2+1)/2+1:n2)
      cdata(:,n2-(n2+1)/2+1:n2)  = temp2d(:,1:(n2+1)/2)
    endif
    cdata = sign1*sign2*cdata/sqrt(1.*n1*n2)
  end subroutine
end module
