MODULE vshot_mod

implicit none

integer, private :: first_cdp 
integer, private :: last_cdp 
integer, private :: zpad 
integer, private :: pad 
integer, private :: src_pad 
integer, private :: nbox 
integer, private :: nz 

real, private :: V_min 
real, private :: V_max 
real, private :: dx 
real, private :: dz 

real, parameter, private :: pi = 3.1415926535897932 ! The Order of Chaos (Darren Aronofsky, 1998)

contains

  subroutine vshot_init(dx_in,dz_in,nz_in,nbox_in,first_cdp_in,last_cdp_in,&
                 pad_in,zpad_in,src_pad_in,V_min_in,V_max_in)

    integer, intent(in) :: nz_in
    integer, intent(in) :: nbox_in
    integer, intent(in) :: first_cdp_in
    integer, intent(in) :: last_cdp_in
    integer, intent(in) :: pad_in
    integer, intent(in) :: zpad_in
    integer, intent(in) :: src_pad_in

    real, intent(in) :: dx_in
    real, intent(in) :: dz_in
    real, intent(in) :: V_min_in
    real, intent(in) :: V_max_in

    nz = nz_in
    nbox = nbox_in
    first_cdp = first_cdp_in
    last_cdp = last_cdp_in
    pad = pad_in
    zpad = zpad_in
    src_pad = src_pad_in
    dx = dx_in
    dz = dz_in
    V_min = V_min_in
    V_max = V_max_in
  end subroutine
  
  subroutine vshot(sgx,V_int,V_shot,rand_bnd,anis)
  !*******************************************************************************
  !*** This subroutine prepares the velocities and anisotropies by shot gather ***
  !*******************************************************************************
    implicit none
    
    integer :: jx
    integer :: jz
    integer :: jgx
    integer :: i 
    integer :: k
    integer :: values(1:8) 
!    integer :: clock 
    
    integer, dimension(:), allocatable :: seed
    
    real :: rand1 
    real :: rand2 
    real :: fact 
    real :: damp1 
    real :: damp2 
    real :: frac 
    real :: Vtest
    real :: distMax 
    real :: distMaxU 
    
    real,dimension(:),allocatable :: distXL
    real,dimension(:),allocatable :: distXR
    real,dimension(:),allocatable :: distZU
    real,dimension(:),allocatable :: distZD
    
    real,dimension(:),allocatable :: dist
    real,dimension(:),allocatable :: distU
    
!    real, dimension(nbox+src_pad), intent(in) :: sgx
    real, dimension(:) :: sgx
    real, dimension(nz,first_cdp:last_cdp),    intent(in) :: V_int
    real, dimension(1-pad-zpad:nz+pad,&
                    1-pad:nbox+pad+src_pad),    intent(out) :: V_shot
    
    logical, intent(in) :: rand_bnd, anis
    V_shot = 0
    if(rand_bnd) then
      fact = dx/dz
      if(fact < 1.0) then
        fact = 1.25/fact
      else
        fact = 1.25*fact
      end if
    else
      fact = 0
    end if 
    
!    call random_seed(size = n)
    call date_and_time(values=values)
    call random_seed(size=k)
    allocate(seed(1:k))
    seed(:) = values(8)
!    call system_clock(count=clock)
!    seed = clock+37*(/(i-1,i = 1,n)/)
    call random_seed(put = seed)
    
    allocate(distZU(1-pad-zpad:0), & !-zpad),   &
             distZD(1:pad),   &
             distXL(1-pad:0),   &
             distXR(1:pad),   & 
             distU(1-pad-zpad:0), &
             dist(1:pad+zpad))
    
    distZU = 0
    distZD = 0
    distXL = 0 
    distXR = 0
    
    damp1 = 1
    damp2 = 1
    frac = 1.0 
    
    do jz =1-pad-zpad,0,1 !-zpad,1
!      distZU(jz) = fact*dz*(1-jz-zpad)
      distZU(jz) = fact*dz*(-jz)
!      damp1 = 1-frac*dist(1-jz-zpad)/distMax
!      distZUnorm(jz) = distZU(jz)/distMax
    end do 
    
    do jz =1,pad
      distZD(jz) = fact*dz*jz
!      distZDnorm(jz) = distZD(jz)/distMax
    end do 
    
    do jx =1-pad,0,1
      distXL(jx) = fact*dx*(1-jx)
!      distXLnorm(jx) = distXL(jx)/distMax
    end do 
    
    do jx =1,pad
      distXR(jx) = fact*dx*jx
!      distXRnorm(jx) = distXR(jx)/distMax
    end do 
    distMax = pad 
    distMaxU = (pad+zpad) 
    do jz =1-pad-zpad,0,1
      distU(jz) = -1.0*jz
    end do 
    do jz =1,pad+zpad
      dist(jz) = 1.0*(jz-1)
    end do 
    
    do jgx = 1-pad,nbox+pad+src_pad
      if(jgx >= 1 .and. jgx <= nbox+src_pad) then
        jx = 1+floor(sgx(jgx)/dx)
        if(jx >= first_cdp .and. jx <= last_cdp) then
          V_shot(1:nz,jgx) = V_int(1:nz,jx)
          do jz = 1-zpad,0
            V_shot(jz,jgx) = V_int(1,jx)
          end do
          do jz = 0,1-pad-zpad,-1
            do 
              rand1 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = V_int(1,jx)*damp1+rand1*distZU(jz)
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = (V_int(nz,jx)*damp1+rand1*distZD(jz-nz))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        elseif(jx < first_cdp) then
          V_shot(1:nz,jgx) = V_int(1:nz,first_cdp)
          do jz = 1-zpad,0
            V_shot(jz,jgx) = V_int(1,first_cdp)
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = (V_int(1,first_cdp)*damp1+rand1*distZU(jz))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = (V_int(nz,first_cdp)*damp1+rand1*distZD(jz-nz))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        elseif(jx > last_cdp) then
          V_shot(1:nz,jgx) = V_int(1:nz,last_cdp)
          do jz = 1-zpad,0
            V_shot(jz,jgx) = V_int(1,last_cdp)
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = (V_int(1,last_cdp)*damp1+rand1*distZU(jz))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                rand1 = 2*(rand1-0.5) 
                damp1 = sin(0.5*pi*damp1)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
              else
                damp1 = sin(0.5*pi*damp1)**2
              end if
              Vtest = (V_int(nz,last_cdp)*damp1+rand1*distZD(jz-nz))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        end if
      elseif(jgx < 1) then
        jx = 1+floor(sgx(1)/dx)
        if(jx >= first_cdp .and. jx <= last_cdp) then
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand2)
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,jx)*damp2+rand2*distXL(jgx))
              else
                Vtest = (V_int(jz,jx)*damp2+rand2*distXL(jgx))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,jx)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,jx)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                 V_shot(jz,jgx) = Vtest
                 exit
              end if
            end do
          end do
        elseif(jx < first_cdp) then
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand2)  !! ALERT ***
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,first_cdp)*damp2+rand2*distXL(jgx))
              else
                Vtest = (V_int(jz,first_cdp)*damp2+rand2*distXL(jgx))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 =frac*(1-distU(jz)/distMaxU)
              damp2 =frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,first_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,first_cdp)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        elseif(jx > last_cdp) then
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand2)
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,last_cdp)*damp2+rand2*distXL(jgx))
              else
                Vtest = (V_int(jz,last_cdp)*damp2+rand2*distXL(jgx))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,last_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(1-jgx)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,last_cdp)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXL(jgx))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        end if
      else
        jx = 1+floor(sgx(nbox+src_pad)/dx)
        if(jx >= first_cdp .and. jx <= last_cdp) then
!write(0,*) 'jx, first_cdp, last_cdp ',jx, first_cdp, last_cdp
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand2)
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,jx)*damp2+rand2*distXR(jgx-nbox-src_pad))
              else
                Vtest = (V_int(jz,jx)*damp2+rand2*distXR(jgx-nbox-src_pad))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,jx)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,jx)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        elseif(jx < first_cdp) then
!write(0,*) 'jx, first_cdp, last_cdp ',jx, first_cdp, last_cdp
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand2)
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,first_cdp)*damp2+rand2*distXR(jgx-nbox-src_pad))
              else
                Vtest = (V_int(jz,first_cdp)*damp2+rand2*distXR(jgx-nbox-src_pad))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,first_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,first_cdp)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        elseif(jx > last_cdp) then
!write(0,*) 'jx, first_cdp, last_cdp ',jx, first_cdp, last_cdp
          do jz = -zpad+1,nz
            do
              rand2 = 0
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand2)
                rand2 = 2*(rand2-0.5) 
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp2 = sin(0.5*pi*damp2)**2
              end if
              if(jz < 1) then
                Vtest = (V_int(1,last_cdp)*damp2+rand2*distXR(jgx-nbox-src_pad))
              else
                Vtest = (V_int(jz,last_cdp)*damp2+rand2*distXR(jgx-nbox-src_pad))
              end if
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
!          do jz = -zpad,1-pad-zpad,-1
          do jz = 0,1-pad-zpad,-1
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-distU(jz)/distMaxU)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(1,last_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
          do jz = nz+1,nz+pad
            do
              rand1 = 0
              rand2 = 0
              damp1 = frac*(1-dist(jz-nz)/distMax)
              damp2 = frac*(1-dist(jgx-nbox-src_pad)/distMax)
              if(rand_bnd) then 
                call random_number(rand1)
                call random_number(rand2)
                rand1 = 2*(rand1-0.5) 
                rand2 = 2*(rand2-0.5) 
                damp1 = sin(0.5*pi*damp1)
                damp2 = sin(0.5*pi*damp2)
              else if(anis) then
                damp1 = 1+cos(0.5*pi*damp1)
                damp2 = 1+cos(0.5*pi*damp2)
              else
                damp1 = sin(0.5*pi*damp1)**2
                damp2 = sin(0.5*pi*damp2)**2
              end if
              Vtest = (V_int(nz,last_cdp)*damp1*damp2+rand1*distZD(jz-nz)+rand2*distXR(jgx-nbox-src_pad))
              if(Vtest <= V_max .and. Vtest > V_min .or. (.not.rand_bnd)) then
                V_shot(jz,jgx) = Vtest
                exit
              end if
            end do
          end do
        end if
      end if
    end do
  end subroutine

  subroutine smooth2d(nz,nx,smooth_x,smooth_z,nsmooth_x,nsmooth_z,Wsmooth_x,Wsmooth_z,v)
  !*****************************************************************************
  !*** Smoothter for the velocity field. Courtesy of Bo Zhang, alias "El Bo" ***
  !*****************************************************************************
    implicit none
  
    integer :: ix ! Inline index
    integer :: iz ! Depth/Time index
    integer :: ik ! Smoothing window index
    
    integer, intent(in) :: nx        ! Number of inlines
    integer, intent(in) :: nz        ! Number of depth/time samples
    integer, intent(in) :: nsmooth_x ! Number of samples in the horizontal smoothing window
    integer, intent(in) :: nsmooth_z ! Number of samples in the vertical smoothing window
    
    real, intent(in) :: Wsmooth_x ! Horizontal smooth window
    real, intent(in) :: Wsmooth_z ! Vertical smooth window
    
    real, dimension(nz,nx), intent(inout) :: v ! Input velocity / output smoothed velocity
    
    logical,intent(in) :: smooth_x ! True if horizontal smooth is required
    logical,intent(in) :: smooth_z ! True if vertical smooth is required
    
    !*************************
    !*** Horizontal smooth ***
    !*************************
    if (smooth_x) then
      do ik = 1,nsmooth_x
        do ix = 2,nx-1
          v(:,ix) = 0.5*(1-Wsmooth_x)*v(:,ix+1)+Wsmooth_x*v(:,ix)+0.5*(1-Wsmooth_x)*v(:,ix-1)
        end do
      end do
    end if
    !***********************
    !*** Vertical smooth ***
    !***********************
    if (smooth_z) then
      do ik = 1,nsmooth_z
        do iz = 2,nz-1
          v(iz,:) = 0.5*(1-Wsmooth_z)*v(iz+1,:)+Wsmooth_z*v(iz,:)+0.5*(1-Wsmooth_z)*v(iz-1,:)
        end do
      end do
    end if
  end subroutine
end module
