module weqsolver_mod 
use sep
!use omp_lib
!use fourier
use lap_c_mod
implicit none
integer, private :: jgx, jx, jz
integer, private :: nB, nt, nts, ngx, nkgx, nz, nkz
integer, private :: nsx_enc, ngx_enc
integer, private :: pad, zshift 

integer, dimension(:), allocatable, private :: encod_vector

real, parameter, private :: pi = 3.1415926535897932 ! The Order of Chaos (Darren Aronofsky, 1998)
real, parameter, private :: tiny = 0.00000001 

real, private, pointer :: zxtaper(:),ctaper(:,:)

real, private :: dt, dx, dz
real, private :: a0
real, private :: a1x, a2x, a3x, a4x, a5x, a6x
real, private :: a1z, a2z, a3z, a4z, a5z, a6z
real, private :: b0x, b0z
real, private :: b1x, b2x, b3x, b4x,b5x,b6x
real, private :: b1z, b2z, b3z, b4z,b5z,b6z

real, dimension(:,:), allocatable, private :: kfact, kfact1, kfact2, kfact3, kfact4, kfact5 
real, dimension(:,:), allocatable, private :: Vshot, Vshotr 
real, dimension(:,:), allocatable, private :: Vanis1, Vanis2, Vanis3, Vanis4, Vanis5 
real, dimension(:,:), allocatable, private :: sinth, costh, sin2th
real, dimension(:),   allocatable, private :: fkzsep, paux, pauxgrad
real, dimension(:,:,:), allocatable, private :: pft_aux
real, dimension(:,:), allocatable, private :: qft

real, dimension(:,:), allocatable, private :: dpx_cur, dpz_cur, drx_cur, drz_cur, dpxz_cur, drxz_cur
real, dimension(:,:), allocatable, private :: dpxz_act, drxz_act, dpxz_new, drxz_new
real, dimension(:,:), allocatable, private :: dpxz_a, drxz_a, pxz, rxz
real, target, dimension(:), allocatable, private :: pcur, rcur, pxcur, pzcur, rxcur, rzcur, pxxcur, pzzcur, rxxcur, rzzcur
real, target, dimension(:), allocatable, private :: pnew, rnew, pxnew, pznew, rxnew, rznew, pxxnew, pzznew, rxxnew, rzznew
real, pointer, dimension(:), private :: ppcur, prcur, ppxcur, ppzcur, prxcur, przcur, ppxxcur, ppzzcur, prxxcur, przzcur
real, pointer, dimension(:), private :: ppnew, prnew, ppxnew, ppznew, prxnew, prznew, ppxxnew, ppzznew, prxxnew, przznew

real, dimension(:),   allocatable, private :: am

real, dimension(:,:,:), allocatable, private :: dpxz, drxz
complex, dimension(:,:), allocatable, private :: dpxz_aux, dpxz_aux2 

logical, private :: eps_f,del_f,th_f,anis,energy

contains
  subroutine param_init(dt_in,dx_in,dz_in,nt_in,nz_in)
    integer, intent(in) :: nz_in, nt_in
    real, intent(in)    :: dt_in, dx_in, dz_in
    nz = nz_in
    nt = nt_in
    dt = dt_in
    dx = dx_in
    dz = dz_in
  end subroutine
  
  subroutine sourc_interp(dt,nt_data,nt_fact,nt,src,s)
  !*** Interpolates the source ***
    integer :: j0
    integer :: j1

    integer, intent(in) :: nt, nt_data

    real :: ft 

    real, intent(in) :: nt_fact, dt

    real, allocatable, dimension(:) :: saux

    real, dimension(:) :: src
    real, dimension(:) :: s

    s=0
! Go for interpolation
    do j0 = 1,size(src)-1
      do j1 = 1,nt_fact
        ft = real(j1-1)/REAL(nt_fact)
        s(j1+(j0-1)*nt_fact) = (1.0-ft)*src(j0)+ft*src(j0+1)
      end do
    end do
    allocate(saux(size(s)))
!    saux = s*0.5*dt**2
!    do j0 = 2,nts
!      s(j0) = saux(j0)+saux(j0-1)
!    end do
  end subroutine

  subroutine rcv_interp(nt_data,nt_fact,nt,rcv,r,adj)
  !*** Interpolates shot traces ***
    integer :: j0
    integer :: j1

    integer, intent(in) :: nt, nt_data
    real, intent(in) :: nt_fact

    real :: ft 

    real, dimension(:)  :: r, rcv 

    logical, intent(in) :: adj 

! Go for interpolation
    if(adj) then
      rcv = 0
    else
      r = 0
    end if
    do j0 = 1,size(rcv)-1
      do j1 = 1,nt_fact
        ft = real(j1-1)/REAL(nt_fact)
        if(adj) then
          rcv(j0) = (1.0-ft)*r(j1+(j0-1)*nt_fact)
          rcv(j0+1) = ft*r(j1+(j0-1)*nt_fact)
        else
          r(j1+(j0-1)*nt_fact) = (1.0-ft)*rcv(j0)+ft*rcv(j0+1)
        end if
      end do
    end do
  end subroutine

  subroutine weqsolver_init(nB_in,nkz_in,nkgx_in,&
                ngx_in,pad_in,anis_in,sblnd,am_in,energy_in,&
                nsx_enc_in,ngx_enc_in,ntraces,nphase,encod_vector_in)
    optional :: nsx_enc_in,ngx_enc_in,ntraces,nphase,encod_vector_in

    integer, intent(in) :: ngx_in, nkgx_in, nkz_in, nB_in, pad_in
    integer, intent(in) :: nsx_enc_in, ngx_enc_in, nphase
    integer, intent(in) :: ntraces

    integer, dimension(:), intent(in) :: encod_vector_in

    real, dimension(:),   intent(in)  :: am_in

    logical, intent(in) :: anis_in, energy_in, sblnd

    nkgx = nkgx_in
    nkz = nkz_in
    ngx = ngx_in
    nB = nB_in
    pad = pad_in
    anis = anis_in
    energy = energy_in

    allocate(am(0:nB))
    am = am_in
    allocate(qft(nkz,nkgx))
    allocate(dpxz_act(nkz,nkgx),&
             dpxz_new(nkz,nkgx))
    allocate(dpxz(nkz,nkgx,nB))
    if(anis) then
      allocate(drxz_act(nkz,nkgx),&
               drxz_new(nkz,nkgx))
      allocate(drxz(nkz,nkgx,nB))
    end if
    if(sblnd) then
      allocate(encod_vector(ntraces))
      allocate(pft_aux(nkz,nkgx,nt))
      encod_vector = encod_vector_in
      nsx_enc = nsx_enc_in
      ngx_enc = ngx_enc_in
    end if
  end subroutine

  subroutine grad_init(ispc)
    implicit none 
    real :: idx, idz

    logical, intent(in) :: ispc

    idx = 1.0/dx
    idz = 1.0/dz

    allocate(dpx_cur(nkz,nkgx))
    allocate(dpz_cur(nkz,nkgx))
    allocate(pauxgrad(MAX(nkz,nkgx)))

    if(ispc) then
      allocate(pxcur(nkz*nkgx))
      allocate(rzcur(nkz*nkgx))
      allocate(pxnew(nkz*nkgx))
      allocate(rznew(nkz*nkgx))
    end if
!*** 12th order approx. table by Fornberg and Ghrist (1999)
    b1x = 160083.0/131072.0 
    b2x = -12705.0/131072.0
    b3x = 22869.0/1310720.0
    b4x = -5445.0/1835008.0
    b5x = 847.0/2359296.0
    b6x = -63.0/2883584.0
    b1z = b1x
    b2z = b2x
    b3z = b3x
    b4z = b4x
    b5z = b5x
    b6z = b6x
    b1x = b1x*idx
    b2x = b2x*idx
    b3x = b3x*idx
    b4x = b4x*idx
    b5x = b5x*idx
    b6x = b6x*idx
    b1z = b1z*idz
    b2z = b2z*idz
    b3z = b3z*idz
    b4z = b4z*idz
    b5z = b5z*idz
    b6z = b6z*idz
  end subroutine

  subroutine lap_init(Vshot_in,Vshotr_in,ispc,&
                sinth_in,costh_in) !,sin2th_in)
    implicit none 
    optional :: sinth_in,costh_in !,sin2th_in
    real*8  :: idx, idz

    real, dimension(:,:) :: Vshot_in, Vshotr_in, sinth_in, costh_in 
!    real, dimension(:,:), intent(in) :: sin2th_in 

    logical, intent(in) :: ispc

    idx = 1.0/(dx*dx)
    idz = 1.0/(dz*dz)

    if(.not. allocateD(Vshot)) then  
      allocate(Vshot(nkz,nkgx))
      Vshot = Vshot_in
    end if
    if(.not. allocateD(Vshotr)) then  
      allocate(Vshotr(nkz,nkgx))
      Vshotr = Vshotr_in
    end if

    if(ispc) then
      allocate(pxxcur(nkz*nkgx))
      allocate(pzzcur(nkz*nkgx))
      allocate(pxxnew(nkz*nkgx))
      allocate(pzznew(nkz*nkgx))
      allocate(rxxcur(nkz*nkgx))
      allocate(rzzcur(nkz*nkgx))
      allocate(rxxnew(nkz*nkgx))
      allocate(rzznew(nkz*nkgx))
      if(anis) then
        allocate(drxz_cur(nkz,nkgx))
        if(PRESENT(sinth_in)) then
          allocate(sinth(nkz,nkgx))
          allocate(costh(nkz,nkgx))
          sinth = sinth_in
          costh = costh_in
          allocate(pxz(nkz,nkgx))
          allocate(rxz(nkz,nkgx))
          if(.not. allocateD(paux)) ALLOCATE(paux(MAX(nkz,nkgx)))
        end if
      end if
    else
      allocate(dpxz_cur(nkz,nkgx))
      if(anis) then
        allocate(drxz_cur(nkz,nkgx))
        if(PRESENT(sinth_in)) then
          allocate(sinth(nkz,nkgx))
          allocate(costh(nkz,nkgx))
          sinth = sinth_in
          costh = costh_in
          allocate(pxz(nkz,nkgx))
          allocate(rxz(nkz,nkgx))
          if(.not. allocateD(paux)) ALLOCATE(paux(MAX(nkz,nkgx)))
          allocate(dpxz_a(nkz,nkgx))
          allocate(drxz_a(nkz,nkgx))
        end if
      end if
    end if

!*** 12th order approx. Zhang and Yao (2013)
    a0 = -3.12513824
    a1x = 1.84108651
    a2x = -0.35706478
    a3x = 0.10185626
    a4x = -0.02924772
    a5x = 0.00696837
    a6x = -0.00102952
    a1z = a1x
    a2z = a2x
    a3z = a3x
    a4z = a4x
    a5z = a5x
    a6z = a6x
    a0  = a0 *(idx+idz)
    a1x = a1x*idx
    a2x = a2x*idx
    a3x = a3x*idx
    a4x = a4x*idx
    a5x = a5x*idx
    a6x = a6x*idx
    a1z = a1z*idz
    a2z = a2z*idz
    a3z = a3z*idz
    a4z = a4z*idz
    a5z = a5z*idz
    a6z = a6z*idz
    call init_taperCos(pad)
  end subroutine

  subroutine xcorr_init(zshift_in,fkzsep_in)
    implicit none
    integer, intent(in) :: zshift_in 
    
    real, dimension(:), intent(in) :: fkzsep_in
    
    allocate(fkzsep(nkz))

    zshift = zshift_in
    fkzsep = fkzsep_in
  end subroutine

  subroutine grad(px,rz,ispc)
  !*** This subroutine applies the factor -L**2 to     ***
  !*** the wavefield using the rapid expansion method  ***
    implicit none
    integer :: n, n1, n2 
  
    real, dimension(:,:) :: px, rz

    logical, intent(in) :: ispc
    dpx_cur = px
    dpz_cur = rz
    px = 0
    rz = 0
    n1 = size(px,1)
    n2 = size(px,2)

    ppxcur => pxcur
    ppxnew => pxnew
    ppzcur => rzcur
    ppznew => rznew
    if(ispc) then
      pxcur = reshape(dpx_cur(1:n1,1:n2),(/n1*n2/))
      rzcur = reshape(dpz_cur(1:n1,1:n2),(/n1*n2/))
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 7,nkgx-5
          call gradXceic(0, 0, ppxcur, ppxnew, nkz, jgx-1, dx)
        end do
!$omp end parallel do
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 1,nkgx
          call gradZceic(0, 0, ppzcur, ppznew, nkz, jgx-1, dz)
        end do
!$omp end parallel do
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!      do jgx = 6,n2-6
!        call gradc(ppxcur,ppxnew,ppzcur,ppznew,n1,n2,jgx-1,dz,dx)
!      end do
!!$omp end parallel do
      px = reshape(pxnew,(/n1,n2/)) 
      rz = reshape(rznew,(/n1,n2/)) 
    else
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
      do jgx = 6,n2-6
        do jz = 1,n1
          px(jz,jgx) = b1x*(dpx_cur(jz,jgx+1)-dpx_cur(jz,jgx  ))+&
                       b2x*(dpx_cur(jz,jgx+2)-dpx_cur(jz,jgx-1))+&
                       b3x*(dpx_cur(jz,jgx+3)-dpx_cur(jz,jgx-2))+&
                       b4x*(dpx_cur(jz,jgx+4)-dpx_cur(jz,jgx-3))+&
                       b5x*(dpx_cur(jz,jgx+5)-dpx_cur(jz,jgx-4))+&
                       b6x*(dpx_cur(jz,jgx+6)-dpx_cur(jz,jgx-5))
        end do
      end do
!$omp end parallel do
    
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
      do jgx = 1,n2
        do jz = 6,n1-6
          rz(jz,jgx) = b1z*(dpz_cur(jz+1,jgx)-dpz_cur(jz,jgx  ))+&
                       b2z*(dpz_cur(jz+2,jgx)-dpz_cur(jz-1,jgx))+&
                       b3z*(dpz_cur(jz+3,jgx)-dpz_cur(jz-2,jgx))+&
                       b4z*(dpz_cur(jz+4,jgx)-dpz_cur(jz-3,jgx))+&
                       b5z*(dpz_cur(jz+5,jgx)-dpz_cur(jz-4,jgx))+&
                       b6z*(dpz_cur(jz+6,jgx)-dpz_cur(jz-5,jgx))
        end do
      end do
!$omp end parallel do
    end if
    pauxgrad=0
!$omp parallel do default(shared) &
!$omp private(pauxgrad,jgx) 
    do jgx = 1,n2
      call interp1d_lop2(.false., .false., rz(:,jgx),pauxgrad(1:n1))
      rz(:,jgx)=pauxgrad(1:n1)
    end do
!$omp end parallel do
    pauxgrad=0
!$omp parallel do default(shared) &
!$omp private(pauxgrad,jz) 
    do jz = 1,n1
      call interp1d_lop2(.false., .false., px(jz,:),pauxgrad(1:n2))
      px(jz,:)=pauxgrad(1:n2)
    end do
!$omp end parallel do
  end subroutine

  subroutine lap_vti(fanis1,fanis2,ispc,isSrc)
  !*** This subroutine applies the factor -L**2 to     ***
  !*** the wavefield using the rapid expansion method  ***
    implicit none
    integer :: jpol
  
    real :: dp_xx, dr_zz

    real, dimension(:,:) :: fanis1, fanis2

    logical, intent(in) :: isSrc, ispc

    ppcur => pcur
    ppnew => pnew
    prcur => rcur
    prnew => rnew
    do jpol = 1,nB
      if(ispc) then
        pcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
        rcur = reshape(drxz(:,:,jpol),(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-6
!          call props_vti(ppcur,ppnew,prcur,prnew,nkz,nkgx,jgx-1,dz,dx)
!        end do
!!$omp end parallel do
!        dpxz_act(:,:) = reshape(pnew,(/nkz,nkgx/))
!        drxz_act(:,:) = reshape(rnew,(/nkz,nkgx/))
        dpxz_act(:,:) = fanis1*reshape(pnew,(/nkz,nkgx/))+fanis2*reshape(rnew,(/nkz,nkgx/))
        drxz_act(:,:) = fanis2*reshape(pnew,(/nkz,nkgx/))+reshape(rnew,(/nkz,nkgx/))
      else
        dpxz_cur = dpxz(:,:,jpol)
        dpxz_act = 0
        drxz_cur = drxz(:,:,jpol)
        drxz_act = 0
!$omp parallel do default(shared) &
!$omp private(jgx,jz,dp_xx,dr_zz) 
        do jgx = 7,nkgx-6
          do jz = 7,nkz-6
            dp_xx = 0.5*a0*dpxz_cur(jz,jgx)+  &
                    a1x*(dpxz_cur(jz,jgx+1)+dpxz_cur(jz,jgx-1))+&
                    a2x*(dpxz_cur(jz,jgx+2)+dpxz_cur(jz,jgx-2))+&
                    a3x*(dpxz_cur(jz,jgx+3)+dpxz_cur(jz,jgx-3))+&
                    a4x*(dpxz_cur(jz,jgx+4)+dpxz_cur(jz,jgx-4))+&
                    a5x*(dpxz_cur(jz,jgx+5)+dpxz_cur(jz,jgx-5))+&
                    a6x*(dpxz_cur(jz,jgx+6)+dpxz_cur(jz,jgx-6))
      
            dr_zz = 0.5*a0*drxz_cur(jz,jgx)+  &
                    a1z*(drxz_cur(jz+1,jgx)+drxz_cur(jz-1,jgx))+&
                    a2z*(drxz_cur(jz+2,jgx)+drxz_cur(jz-2,jgx))+&
                    a3z*(drxz_cur(jz+3,jgx)+drxz_cur(jz-3,jgx))+&
                    a4z*(drxz_cur(jz+4,jgx)+drxz_cur(jz-4,jgx))+&
                    a5z*(drxz_cur(jz+5,jgx)+drxz_cur(jz-5,jgx))+&
                    a6z*(drxz_cur(jz+6,jgx)+drxz_cur(jz-6,jgx))
            dpxz_act(jz,jgx) = fanis1(jz,jgx)*dp_xx+fanis2(jz,jgx)*dr_zz
            drxz_act(jz,jgx) = fanis2(jz,jgx)*dp_xx+dr_zz
          end do
        end do
!$omp end parallel do
      end if
      if(isSrc) then
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshot(jz,jgx)**2
            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshot(jz,jgx)**2
          end do
        end do
!$omp end parallel do
      else
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshotr(jz,jgx)**2
            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshotr(jz,jgx)**2
          end do
        end do
!$omp end parallel do
      end if
      if(jpol == nB) EXIT
      dpxz(:,:,jpol+1) = dpxz(:,:,jpol)
      drxz(:,:,jpol+1) = drxz(:,:,jpol)
    end do
  end subroutine

  subroutine lap_tti(fanis1,fanis2,ispc,isSrc)
  !*** This subroutine applies the factor -L**2 to     ***
  !*** the wavefield using the rapid expansion method  ***
    implicit none
    integer :: jpol, n
  
    real :: dp_xx, dp_zz, dr_xx, dr_zz
  
    real, dimension(:,:),intent(in) :: fanis1, fanis2

    logical, intent(in) :: ispc, isSrc 

    ppxxcur => pxxcur
    ppxxnew => pxxnew
    ppzzcur => pzzcur
    ppzznew => pzznew
    prxxcur => rxxcur
    prxxnew => rxxnew
    przzcur => rzzcur
    przznew => rzznew
    do jpol = 1,nB
      if(ispc) then
        dpxz_act = 0
        drxz_act = 0
        pxxcur = reshape(dpxz(:,:,jpol)*costh,(/nkz*nkgx/))
        pzzcur = reshape(dpxz(:,:,jpol)*sinth,(/nkz*nkgx/))
        rxxcur = reshape(drxz(:,:,jpol)*sinth,(/nkz*nkgx/))
        rzzcur = reshape(drxz(:,:,jpol)*costh,(/nkz*nkgx/))
!************ Apply second derivatives cascading first derivatives and sinusoidal functions
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 7,nkgx-5
          call gradX2c(0, 0, ppxxcur, ppxxnew, nkz, jgx-1, dx)
        end do
!$omp end parallel do
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 1,nkgx
          call gradZ2c(0, 0, ppzzcur, ppzznew, nkz, jgx-1, dz)
        end do
!$omp end parallel do

!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 7,nkgx-5
          call gradX2c(0, 0, prxxcur, prxxnew, nkz, jgx-1, dx)
        end do
!$omp end parallel do
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 1,nkgx
          call gradZ2c(0, 0, przzcur, przznew, nkz, jgx-1, dz)
        end do
!$omp end parallel do

        dpxz_act(:,:) = dpxz_act(:,:)+fanis1*reshape(pxxnew,(/nkz,nkgx/))*costh&
                                     +fanis2*reshape(rxxnew,(/nkz,nkgx/))*sinth
        drxz_act(:,:) = drxz_act(:,:)+fanis2*reshape(pxxnew,(/nkz,nkgx/))*costh&
                                     +       reshape(rxxnew,(/nkz,nkgx/))*sinth

        dpxz_act(:,:) = dpxz_act(:,:)+fanis1*reshape(pzznew,(/nkz,nkgx/))*sinth&
                                     +fanis2*reshape(rzznew,(/nkz,nkgx/))*costh
        drxz_act(:,:) = drxz_act(:,:)+fanis2*reshape(pzznew,(/nkz,nkgx/))*sinth&
                                     +       reshape(rzznew,(/nkz,nkgx/))*costh
!********** First pass cross derivative
        pzzcur = reshape(dpxz(:,:,jpol)*sinth,(/nkz*nkgx/))
        rzzcur = reshape(drxz(:,:,jpol)*costh,(/nkz*nkgx/))
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 1,nkgx
          call gradZc(0, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
        end do
!$omp end parallel do

        pxz = reshape(pzznew,(/nkz,nkgx/))
        rxz = reshape(rzznew,(/nkz,nkgx/))
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jgx) 
        do jgx = 1,nkgx
          call interp1d_lop2(.false., .false., rxz(:,jgx),paux(1:nkz))
          rxz(:,jgx)=paux(1:nkz)
        end do
!$omp end parallel do
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jgx) 
        do jgx = 1,nkgx
          call interp1d_lop2(.false., .false., pxz(:,jgx),paux(1:nkz))
          pxz(:,jgx)=paux(1:nkz)
        end do
!$omp end parallel do
!        pxz = pxz*costh
!        rxz = rxz*sinth
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jz) 
        do jz = 1,nkz
          call interp1d_lop2(.true., .false., paux(1:nkgx),pxz(jz,:))
!          call interp1d_lop2(.false., .false., paux(1:nkgx),pxz(jz,:))
          pxz(jz,:)=paux(1:nkgx)
        end do
!$omp end parallel do
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jz) 
        do jz = 1,nkz
          call interp1d_lop2(.true., .false., paux(1:nkgx),rxz(jz,:))
!          call interp1d_lop2(.false., .false., paux(1:nkgx),rxz(jz,:))
          rxz(jz,:)=paux(1:nkgx)
        end do
!$omp end parallel do
        pxxcur = reshape(pxz,(/nkz*nkgx/))
        rxxcur = reshape(rxz,(/nkz*nkgx/))
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 7,nkgx-5
          call gradXc(1, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
        end do
!$omp end parallel do
        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pxxnew,(/nkz,nkgx/))*costh&
                                     +fanis2*reshape(rxxnew,(/nkz,nkgx/))*sinth
        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pxxnew,(/nkz,nkgx/))*costh&
                                     +       reshape(rxxnew,(/nkz,nkgx/))*sinth
!********** Second pass cross derivative
        pxxcur = reshape(dpxz(:,:,jpol)*costh,(/nkz*nkgx/))
        rxxcur = reshape(drxz(:,:,jpol)*sinth,(/nkz*nkgx/))
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 7,nkgx-5
          call gradXc(0, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
        end do
!$omp end parallel do
        pxz = reshape(ppxxnew,(/nkz,nkgx/))
        rxz = reshape(prxxnew,(/nkz,nkgx/))
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jz) 
        do jz = 1,nkz
          call interp1d_lop2(.false., .false., rxz(jz,:),paux(1:nkgx))
          rxz(jz,:)=paux(1:nkgx)
        end do
!$omp end parallel do
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jz) 
        do jz = 1,nkz
          call interp1d_lop2(.false., .false., pxz(jz,:),paux(1:nkgx))
          pxz(jz,:)=paux(1:nkgx)
        end do
!$omp end parallel do

!        pxz = pxz*sinth
!        rxz = rxz*costh
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jgx) 
        do jgx = 1,nkgx
          call interp1d_lop2(.true., .false., paux(1:nkz),pxz(:,jgx))
!          call interp1d_lop2(.false., .false., paux(1:nkz),pxz(:,jgx))
          pxz(:,jgx)=paux(1:nkz)
        end do
!$omp end parallel do
        paux=0
!$omp parallel do default(shared) &
!$omp private(paux,jgx) 
        do jgx = 1,nkgx
          call interp1d_lop2(.true., .false., paux(1:nkz),rxz(:,jgx))
!          call interp1d_lop2(.false., .false., paux(1:nkz),rxz(:,jgx))
          rxz(:,jgx)=paux(1:nkz)
        end do
!$omp end parallel do
        pzzcur = reshape(pxz,(/nkz*nkgx/))
        rzzcur = reshape(rxz,(/nkz*nkgx/))
!$omp parallel do default(shared) &
!$omp private(jgx) 
        do jgx = 1,nkgx
          call gradZc(1, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
        end do
!$omp end parallel do
        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pzznew,(/nkz,nkgx/))*sinth&
                                     +fanis2*reshape(rzznew,(/nkz,nkgx/))*costh
        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pzznew,(/nkz,nkgx/))*sinth&
                                     +       reshape(rzznew,(/nkz,nkgx/))*costh
      end if

      if(isSrc) then
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshot(jz,jgx)*Vshot(jz,jgx)
            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshot(jz,jgx)*Vshot(jz,jgx)
          end do
        end do
!$omp end parallel do
      else
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshotr(jz,jgx)*Vshotr(jz,jgx)
            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshotr(jz,jgx)*Vshotr(jz,jgx)
          end do
        end do
!$omp end parallel do
      end if

      if(jpol == nB) EXIT
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
      do jgx = 1,nkgx
        do jz = 1,nkz
          dpxz(jz,jgx,jpol+1) = dpxz(jz,jgx,jpol)
          drxz(jz,jgx,jpol+1) = drxz(jz,jgx,jpol)
        end do
      end do
!$omp end parallel do
    end do
  end subroutine








!  subroutine lap_tti(fanis1,fanis2,ispc,isSrc)
!  !*** This subroutine applies the factor -L**2 to     ***
!  !*** the wavefield using the rapid expansion method  ***
!    implicit none
!    integer :: jpol, n
!  
!    real :: dp_xx, dp_zz, dr_xx, dr_zz
!  
!    real, dimension(:,:),intent(in) :: fanis1, fanis2
!
!    logical, intent(in) :: ispc, isSrc 
!
!    ppxxcur => pxxcur
!    ppxxnew => pxxnew
!    ppzzcur => pzzcur
!    ppzznew => pzznew
!    prxxcur => rxxcur
!    prxxnew => rxxnew
!    przzcur => rzzcur
!    przznew => rzznew
!    do jpol = 1,nB
!      if(ispc) then
!        dpxz_act = 0
!        drxz_act = 0
!        pxxcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
!        pzzcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
!        rxxcur = reshape(drxz(:,:,jpol),(/nkz*nkgx/))
!        rzzcur = reshape(drxz(:,:,jpol),(/nkz*nkgx/))
!!************ Apply second derivatives cascading first derivatives and sinusoidal functions
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-5
!          call gradXc(0, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
!        end do
!!$omp end parallel do
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 1,nkgx
!          call gradZc(0, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
!        end do
!!$omp end parallel do
!        pxz = reshape(pxxnew,(/nkz,nkgx/))
!        rxz = reshape(rxxnew,(/nkz,nkgx/))
!
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., pxz(jz,:),paux(1:nkgx))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., rxz(jz,:),paux(1:nkgx))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!
!!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jz) 
!!        do jz = 1,nkz
!!          call interp1d_lop2(.true., .false., paux(1:nkgx),costh(jz,:))
!!          pxz(jz,:)=pxz(jz,:)*paux(1:nkgx)**2
!!        end do
!!!$omp end parallel do
!!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jz) 
!!        do jz = 1,nkz
!!          call interp1d_lop2(.true., .false., paux(1:nkgx),sinth(jz,:))
!!          rxz(jz,:)=rxz(jz,:)*paux(1:nkgx)**2
!!        end do
!!!$omp end parallel do
!
!        pxz = pxz*costh**2
!        rxz = rxz*sinth**2
!
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),pxz(jz,:))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),rxz(jz,:))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!        pxxcur = reshape(pxz,(/nkz*nkgx/))
!        rxxcur = reshape(rxz,(/nkz*nkgx/))
!
!        pxz = reshape(pzznew,(/nkz,nkgx/))
!        rxz = reshape(rzznew,(/nkz,nkgx/))
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., pxz(:,jgx),paux(1:nkz))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., rxz(:,jgx),paux(1:nkz))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jgx) 
!!        do jgx = 1,nkgx
!!          call interp1d_lop2(.true., .false., paux(1:nkz),sinth(:,jgx))
!!          pxz(:,jgx)=pxz(:,jgx)*paux(1:nkz)**2
!!        end do
!!!$omp end parallel do
!!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jgx) 
!!        do jgx = 1,nkgx
!!          call interp1d_lop2(.true., .false., paux(1:nkz),costh(:,jgx))
!!          rxz(:,jgx)=rxz(:,jgx)*paux(1:nkz)**2
!!        end do
!!!$omp end parallel do
!
!
!        pxz = pxz*sinth**2
!        rxz = rxz*costh**2
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),pxz(:,jgx))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),rxz(:,jgx))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        pzzcur = reshape(pxz,(/nkz*nkgx/))
!        rzzcur = reshape(rxz,(/nkz*nkgx/))
!
!!        pxxcur = pxxnew*reshape(costh(:,:),(/nkz*nkgx/))**2
!!        rxxcur = rxxnew*reshape(sinth(:,:),(/nkz*nkgx/))**2
!!        pzzcur = pzznew*reshape(sinth(:,:),(/nkz*nkgx/))**2
!!        rzzcur = rzznew*reshape(costh(:,:),(/nkz*nkgx/))**2
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-5
!          call gradXc(1, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
!        end do
!!$omp end parallel do
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 1,nkgx
!          call gradZc(1, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
!        end do
!!$omp end parallel do
!        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pxxnew,(/nkz,nkgx/))&
!                                     -fanis2*reshape(rxxnew,(/nkz,nkgx/))
!        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pxxnew,(/nkz,nkgx/))&
!                                     -       reshape(rxxnew,(/nkz,nkgx/))
!
!        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pzznew,(/nkz,nkgx/))&
!                                     -fanis2*reshape(rzznew,(/nkz,nkgx/))
!        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pzznew,(/nkz,nkgx/))&
!                                     -       reshape(rzznew,(/nkz,nkgx/))
!if(.true.) then 
!!********** First pass cross derivative
!        pzzcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
!        rzzcur = reshape(drxz(:,:,jpol),(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 1,nkgx
!          call gradZc(0, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
!        end do
!!$omp end parallel do
!
!        pxz = reshape(pzznew,(/nkz,nkgx/))
!        rxz = reshape(rzznew,(/nkz,nkgx/))
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., rxz(:,jgx),paux(1:nkz))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., pxz(:,jgx),paux(1:nkz))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        pxz = pxz*sinth*costh
!        rxz = rxz*costh*sinth
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),pxz(jz,:))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),rxz(jz,:))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!        pxxcur = reshape(pxz,(/nkz*nkgx/))
!        rxxcur = reshape(rxz,(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-5
!          call gradXc(1, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
!        end do
!!$omp end parallel do
!        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pxxnew,(/nkz,nkgx/))&
!                                     +fanis2*reshape(rxxnew,(/nkz,nkgx/))
!        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pxxnew,(/nkz,nkgx/))&
!                                     +       reshape(rxxnew,(/nkz,nkgx/))
!!********** Second pass cross derivative
!        pxxcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
!        rxxcur = reshape(drxz(:,:,jpol),(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-5
!          call gradXc(0, 0, ppxxcur, ppxxnew, prxxcur, prxxnew, nkz, jgx-1, dx)
!        end do
!!$omp end parallel do
!        pxz = reshape(ppxxnew,(/nkz,nkgx/))
!        rxz = reshape(prxxnew,(/nkz,nkgx/))
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., rxz(jz,:),paux(1:nkgx))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., pxz(jz,:),paux(1:nkgx))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!
!        pxz = pxz*costh*sinth
!        rxz = rxz*sinth*costh
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),pxz(:,jgx))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),rxz(:,jgx))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        pzzcur = reshape(pxz,(/nkz*nkgx/))
!        rzzcur = reshape(rxz,(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 1,nkgx
!          call gradZc(1, 0, ppzzcur, ppzznew, przzcur, przznew, nkz, jgx-1, dz)
!        end do
!!$omp end parallel do
!        dpxz_act(:,:) = dpxz_act(:,:)-fanis1*reshape(pzznew,(/nkz,nkgx/))&
!                                     +fanis2*reshape(rzznew,(/nkz,nkgx/))
!        drxz_act(:,:) = drxz_act(:,:)-fanis2*reshape(pzznew,(/nkz,nkgx/))&
!                                     +       reshape(rzznew,(/nkz,nkgx/))
!end if
!      else
!!********** Implement second derivatives
!        dpxz_cur = dpxz(:,:,jpol)
!        drxz_cur = drxz(:,:,jpol)
!        dpxz_act = 0
!        drxz_act = 0
!!************ Apply second derivatives cascading first derivatives and sinusoidal functions
!dpxz_a = 0
!drxz_a = 0
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz,dp_xx,dr_xx) 
!        do jgx = 6,nkgx-6
!          do jz = 1,nkz
!            dp_xx = b1x*(dpxz_cur(jz,jgx+1)-dpxz_cur(jz,jgx-0))+&
!                    b2x*(dpxz_cur(jz,jgx+2)-dpxz_cur(jz,jgx-1))+&
!                    b3x*(dpxz_cur(jz,jgx+3)-dpxz_cur(jz,jgx-2))+&
!                    b4x*(dpxz_cur(jz,jgx+4)-dpxz_cur(jz,jgx-3))+&
!                    b5x*(dpxz_cur(jz,jgx+5)-dpxz_cur(jz,jgx-4))+&
!                    b6x*(dpxz_cur(jz,jgx+6)-dpxz_cur(jz,jgx-5))
!
!            dr_xx = b1x*(drxz_cur(jz,jgx+1)-drxz_cur(jz,jgx-0))+&
!                    b2x*(drxz_cur(jz,jgx+2)-drxz_cur(jz,jgx-1))+&
!                    b3x*(drxz_cur(jz,jgx+3)-drxz_cur(jz,jgx-2))+&
!                    b4x*(drxz_cur(jz,jgx+4)-drxz_cur(jz,jgx-3))+&
!                    b5x*(drxz_cur(jz,jgx+5)-drxz_cur(jz,jgx-4))+&
!                    b6x*(drxz_cur(jz,jgx+6)-drxz_cur(jz,jgx-5))
!            dpxz_a(jz,jgx) = dpxz_a(jz,jgx)+dp_xx*costh(jz,jgx)**2
!            drxz_a(jz,jgx) = drxz_a(jz,jgx)+dr_xx*sinth(jz,jgx)**2
!          end do
!        end do
!!$omp end parallel do
!        pxz = 0
!        rxz = 0
!!!$omp parallel do default(shared) &
!!!$omp private(jgx,jz) 
!        do jgx = 6,nkgx-6
!          do jz = 1,nkz
!            pxz(jz,jgx+6) = pxz(jz,jgx+6)+b6x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+5) = pxz(jz,jgx+5)+b5x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+4) = pxz(jz,jgx+4)+b4x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+3) = pxz(jz,jgx+3)+b3x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+2) = pxz(jz,jgx+2)+b2x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+1) = pxz(jz,jgx+1)+b1x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-0) = pxz(jz,jgx-0)-b1x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-1) = pxz(jz,jgx-1)-b2x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-2) = pxz(jz,jgx-2)-b3x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-3) = pxz(jz,jgx-3)-b4x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-4) = pxz(jz,jgx-4)-b5x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-5) = pxz(jz,jgx-5)-b6x*dpxz_a(jz,jgx)
!                                        
!            rxz(jz,jgx+6) = rxz(jz,jgx+6)+b6x*drxz_a(jz,jgx)
!            rxz(jz,jgx+5) = rxz(jz,jgx+5)+b5x*drxz_a(jz,jgx)
!            rxz(jz,jgx+4) = rxz(jz,jgx+4)+b4x*drxz_a(jz,jgx)
!            rxz(jz,jgx+3) = rxz(jz,jgx+3)+b3x*drxz_a(jz,jgx)
!            rxz(jz,jgx+2) = rxz(jz,jgx+2)+b2x*drxz_a(jz,jgx)
!            rxz(jz,jgx+1) = rxz(jz,jgx+1)+b1x*drxz_a(jz,jgx)
!            rxz(jz,jgx-0) = rxz(jz,jgx-0)-b1x*drxz_a(jz,jgx)
!            rxz(jz,jgx-1) = rxz(jz,jgx-1)-b2x*drxz_a(jz,jgx)
!            rxz(jz,jgx-2) = rxz(jz,jgx-2)-b3x*drxz_a(jz,jgx)
!            rxz(jz,jgx-3) = rxz(jz,jgx-3)-b4x*drxz_a(jz,jgx)
!            rxz(jz,jgx-4) = rxz(jz,jgx-4)-b5x*drxz_a(jz,jgx)
!            rxz(jz,jgx-5) = rxz(jz,jgx-5)-b6x*drxz_a(jz,jgx)
!          end do
!        end do
!!!$omp end parallel do
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_act(jz,jgx) = dpxz_act(jz,jgx)-pxz(jz,jgx)*fanis1(jz,jgx)& !*costh(jz,jgx)&
!                                               -rxz(jz,jgx)*fanis2(jz,jgx) !*sinth(jz,jgx)
!            drxz_act(jz,jgx) = drxz_act(jz,jgx)-pxz(jz,jgx)*fanis2(jz,jgx)& !*costh(jz,jgx)&
!                                               -rxz(jz,jgx)!*sinth(jz,jgx)
!            if(ISNAN(dpxz_act(jz,jgx)) .OR. ISNAN(drxz_act(jz,jgx))) then
!              write(0,*) '2nd deriv1: NaN in pxz or rxz,jgz,jgx',pxz(jz,jgx),rxz(jz,jgx),jz,jgx
!              stop 666
!            end if
!          end do
!        end do
!!$omp end parallel do
!dpxz_a = 0
!drxz_a = 0
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz,dp_xx,dr_xx) 
!        do jgx = 1,nkgx
!          do jz = 6,nkz-6
!            dp_zz = b1z*(dpxz_cur(jz+1,jgx)-dpxz_cur(jz-0,jgx))+&
!                    b2z*(dpxz_cur(jz+2,jgx)-dpxz_cur(jz-1,jgx))+&
!                    b3z*(dpxz_cur(jz+3,jgx)-dpxz_cur(jz-2,jgx))+&
!                    b4z*(dpxz_cur(jz+4,jgx)-dpxz_cur(jz-3,jgx))+&
!                    b5z*(dpxz_cur(jz+5,jgx)-dpxz_cur(jz-4,jgx))+&
!                    b6z*(dpxz_cur(jz+6,jgx)-dpxz_cur(jz-5,jgx))
!
!            dr_zz = b1z*(drxz_cur(jz+1,jgx)-drxz_cur(jz-0,jgx))+&
!                    b2z*(drxz_cur(jz+2,jgx)-drxz_cur(jz-1,jgx))+&
!                    b3z*(drxz_cur(jz+3,jgx)-drxz_cur(jz-2,jgx))+&
!                    b4z*(drxz_cur(jz+4,jgx)-drxz_cur(jz-3,jgx))+&
!                    b5z*(drxz_cur(jz+5,jgx)-drxz_cur(jz-4,jgx))+&
!                    b6z*(drxz_cur(jz+6,jgx)-drxz_cur(jz-5,jgx))
!            dpxz_a(jz,jgx) = dpxz_a(jz,jgx)+dp_zz*sinth(jz,jgx)**2
!            drxz_a(jz,jgx) = drxz_a(jz,jgx)+dr_zz*costh(jz,jgx)**2
!          end do
!        end do
!!$omp end parallel do
!        pxz = 0
!        rxz = 0
!!!$omp parallel do default(shared) &
!!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 6,nkz-6
!            pxz(jz+6,jgx) = pxz(jz+6,jgx)+b6z*dpxz_a(jz,jgx)
!            pxz(jz+5,jgx) = pxz(jz+5,jgx)+b5z*dpxz_a(jz,jgx)
!            pxz(jz+4,jgx) = pxz(jz+4,jgx)+b4z*dpxz_a(jz,jgx)
!            pxz(jz+3,jgx) = pxz(jz+3,jgx)+b3z*dpxz_a(jz,jgx)
!            pxz(jz+2,jgx) = pxz(jz+2,jgx)+b2z*dpxz_a(jz,jgx)
!            pxz(jz+1,jgx) = pxz(jz+1,jgx)+b1z*dpxz_a(jz,jgx)
!            pxz(jz-0,jgx) = pxz(jz-0,jgx)-b1z*dpxz_a(jz,jgx)
!            pxz(jz-1,jgx) = pxz(jz-1,jgx)-b2z*dpxz_a(jz,jgx)
!            pxz(jz-2,jgx) = pxz(jz-2,jgx)-b3z*dpxz_a(jz,jgx)
!            pxz(jz-3,jgx) = pxz(jz-3,jgx)-b4z*dpxz_a(jz,jgx)
!            pxz(jz-4,jgx) = pxz(jz-4,jgx)-b5z*dpxz_a(jz,jgx)
!            pxz(jz-5,jgx) = pxz(jz-5,jgx)-b6z*dpxz_a(jz,jgx)
!                                    
!            rxz(jz+6,jgx) = rxz(jz+6,jgx)+b6z*drxz_a(jz,jgx)
!            rxz(jz+5,jgx) = rxz(jz+5,jgx)+b5z*drxz_a(jz,jgx)
!            rxz(jz+4,jgx) = rxz(jz+4,jgx)+b4z*drxz_a(jz,jgx)
!            rxz(jz+3,jgx) = rxz(jz+3,jgx)+b3z*drxz_a(jz,jgx)
!            rxz(jz+2,jgx) = rxz(jz+2,jgx)+b2z*drxz_a(jz,jgx)
!            rxz(jz+1,jgx) = rxz(jz+1,jgx)+b1z*drxz_a(jz,jgx)
!            rxz(jz-0,jgx) = rxz(jz-0,jgx)-b1z*drxz_a(jz,jgx)
!            rxz(jz-1,jgx) = rxz(jz-1,jgx)-b2z*drxz_a(jz,jgx)
!            rxz(jz-2,jgx) = rxz(jz-2,jgx)-b3z*drxz_a(jz,jgx)
!            rxz(jz-3,jgx) = rxz(jz-3,jgx)-b4z*drxz_a(jz,jgx)
!            rxz(jz-4,jgx) = rxz(jz-4,jgx)-b5z*drxz_a(jz,jgx)
!            rxz(jz-5,jgx) = rxz(jz-5,jgx)-b6z*drxz_a(jz,jgx)
!          end do
!        end do
!!!$omp end parallel do
!
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_act(jz,jgx) = dpxz_act(jz,jgx)-pxz(jz,jgx)*fanis1(jz,jgx)& !*sinth(jz,jgx)&
!                                               -rxz(jz,jgx)*fanis2(jz,jgx) !*costh(jz,jgx)
!            drxz_act(jz,jgx) = drxz_act(jz,jgx)-pxz(jz,jgx)*fanis2(jz,jgx)& !*sinth(jz,jgx)&
!                                               -rxz(jz,jgx)!*costh(jz,jgx)
!            if(ISNAN(dpxz_act(jz,jgx)) .OR. ISNAN(drxz_act(jz,jgx))) then
!              write(0,*) '2nd deriv2: NaN in pxz or rxz,jgz,jgx',pxz(jz,jgx),rxz(jz,jgx),jz,jgx
!              stop 666
!            end if
!          end do
!        end do
!!$omp end parallel do
!
!!********** First pass cross derivative
!        pxz = 0
!        rxz = 0
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_a(jz,jgx)=dpxz_cur(jz,jgx)!*sinth(jz,jgx)!*costh(jz,jgx)
!            drxz_a(jz,jgx)=drxz_cur(jz,jgx)!*costh(jz,jgx)!*sinth(jz,jgx)
!          end do
!        end do
!!$omp end parallel do
!
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 6,nkz-6
!            pxz(jz,jgx) = b1z*(dpxz_a(jz+1,jgx)-dpxz_a(jz  ,jgx))+&
!                          b2z*(dpxz_a(jz+2,jgx)-dpxz_a(jz-1,jgx))+&
!                          b3z*(dpxz_a(jz+3,jgx)-dpxz_a(jz-2,jgx))+&
!                          b4z*(dpxz_a(jz+4,jgx)-dpxz_a(jz-3,jgx))+&
!                          b5z*(dpxz_a(jz+5,jgx)-dpxz_a(jz-4,jgx))+&
!                          b6z*(dpxz_a(jz+6,jgx)-dpxz_a(jz-5,jgx))
!      
!            rxz(jz,jgx) = b1z*(drxz_a(jz+1,jgx)-drxz_a(jz  ,jgx))+&
!                          b2z*(drxz_a(jz+2,jgx)-drxz_a(jz-1,jgx))+&
!                          b3z*(drxz_a(jz+3,jgx)-drxz_a(jz-2,jgx))+&
!                          b4z*(drxz_a(jz+4,jgx)-drxz_a(jz-3,jgx))+&
!                          b5z*(drxz_a(jz+5,jgx)-drxz_a(jz-4,jgx))+&
!                          b6z*(drxz_a(jz+6,jgx)-drxz_a(jz-5,jgx))
!          end do
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., rxz(:,jgx),paux(1:nkz))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jgx) 
!        do jgx = 1,nkgx
!          call interp1d_lop2(.false., .false., pxz(:,jgx),paux(1:nkz))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!!$omp end parallel do
!        pxz = pxz*sinth*costh
!        rxz = rxz*costh*sinth
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),pxz(jz,:))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!        paux=0
!!$omp parallel do default(shared) &
!!$omp private(paux,jz) 
!        do jz = 1,nkz
!          call interp1d_lop2(.true., .false., paux(1:nkgx),rxz(jz,:))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!!$omp end parallel do
!
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_a(jz,jgx)=pxz(jz,jgx)
!            drxz_a(jz,jgx)=rxz(jz,jgx)
!          end do
!        end do
!!$omp end parallel do
!        pxz=0
!        rxz=0 
!!!$omp parallel do default(shared) &
!!!$omp private(jgx,jz) 
!        do jgx = 6,nkgx-6
!          do jz = 1,nkz
!            pxz(jz,jgx+6) = pxz(jz,jgx+6)+b6x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+5) = pxz(jz,jgx+5)+b5x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+4) = pxz(jz,jgx+4)+b4x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+3) = pxz(jz,jgx+3)+b3x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+2) = pxz(jz,jgx+2)+b2x*dpxz_a(jz,jgx)
!            pxz(jz,jgx+1) = pxz(jz,jgx+1)+b1x*dpxz_a(jz,jgx)
!            pxz(jz,jgx  ) = pxz(jz,jgx  )-b1x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-1) = pxz(jz,jgx-1)-b2x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-2) = pxz(jz,jgx-2)-b3x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-3) = pxz(jz,jgx-3)-b4x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-4) = pxz(jz,jgx-4)-b5x*dpxz_a(jz,jgx)
!            pxz(jz,jgx-5) = pxz(jz,jgx-5)-b6x*dpxz_a(jz,jgx)
!
!            rxz(jz,jgx+6) = rxz(jz,jgx+6)+b6x*drxz_a(jz,jgx)
!            rxz(jz,jgx+5) = rxz(jz,jgx+5)+b5x*drxz_a(jz,jgx)
!            rxz(jz,jgx+4) = rxz(jz,jgx+4)+b4x*drxz_a(jz,jgx)
!            rxz(jz,jgx+3) = rxz(jz,jgx+3)+b3x*drxz_a(jz,jgx)
!            rxz(jz,jgx+2) = rxz(jz,jgx+2)+b2x*drxz_a(jz,jgx)
!            rxz(jz,jgx+1) = rxz(jz,jgx+1)+b1x*drxz_a(jz,jgx)
!            rxz(jz,jgx  ) = rxz(jz,jgx  )-b1x*drxz_a(jz,jgx)
!            rxz(jz,jgx-1) = rxz(jz,jgx-1)-b2x*drxz_a(jz,jgx)
!            rxz(jz,jgx-2) = rxz(jz,jgx-2)-b3x*drxz_a(jz,jgx)
!            rxz(jz,jgx-3) = rxz(jz,jgx-3)-b4x*drxz_a(jz,jgx)
!            rxz(jz,jgx-4) = rxz(jz,jgx-4)-b5x*drxz_a(jz,jgx)
!            rxz(jz,jgx-5) = rxz(jz,jgx-5)-b6x*drxz_a(jz,jgx)
!          end do
!        end do
!!!$omp end parallel do
!
!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jz) 
!!        do jz = 1,nkz
!!          call interp1d_lop2(.false., .false., pxz(jz,:),paux(1:nkgx))
!!          pxz(jz,:)=paux(1:nkgx)
!!        end do
!!!$omp end parallel do
!        paux=0
!!!$omp parallel do default(shared) &
!!!$omp private(paux,jz) 
!!        do jz = 1,nkz
!!          call interp1d_lop2(.false., .false., rxz(jz,:),paux(1:nkgx))
!!          rxz(jz,:)=paux(1:nkgx)
!!        end do
!!!$omp end parallel do
!
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_act(jz,jgx) = dpxz_act(jz,jgx)-1*(pxz(jz,jgx)*fanis1(jz,jgx)&
!                                                  -rxz(jz,jgx)*fanis2(jz,jgx))
!            drxz_act(jz,jgx) = drxz_act(jz,jgx)-1*(pxz(jz,jgx)*fanis2(jz,jgx)&
!                                                  -rxz(jz,jgx))
!            if(ISNAN(dpxz_act(jz,jgx)) .OR. ISNAN(drxz_act(jz,jgx))) then
!              write(0,*) 'NaN in pxz or rxz,jgz,jgx',pxz(jz,jgx),rxz(jz,jgx),jz,jgx
!              stop 666
!            end if
!          end do
!        end do
!!$omp end parallel do
!
!!********** Second pass cross derivative
!        dpxz_a=dpxz_cur!*costh!*sinth
!        drxz_a=drxz_cur!*sinth!*costh
!        pxz = 0
!        rxz = 0
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 6,nkgx-6
!          do jz = 1,nkz
!            pxz(jz,jgx) = b1x*(dpxz_a(jz,jgx+1)-dpxz_a(jz,jgx  ))+&
!                          b2x*(dpxz_a(jz,jgx+2)-dpxz_a(jz,jgx-1))+&
!                          b3x*(dpxz_a(jz,jgx+3)-dpxz_a(jz,jgx-2))+&
!                          b4x*(dpxz_a(jz,jgx+4)-dpxz_a(jz,jgx-3))+&
!                          b5x*(dpxz_a(jz,jgx+5)-dpxz_a(jz,jgx-4))+&
!                          b6x*(dpxz_a(jz,jgx+6)-dpxz_a(jz,jgx-5))
!      
!            rxz(jz,jgx) = b1x*(drxz_a(jz,jgx+1)-drxz_a(jz,jgx  ))+&
!                          b2x*(drxz_a(jz,jgx+2)-drxz_a(jz,jgx-1))+&
!                          b3x*(drxz_a(jz,jgx+3)-drxz_a(jz,jgx-2))+&
!                          b4x*(drxz_a(jz,jgx+4)-drxz_a(jz,jgx-3))+&
!                          b5x*(drxz_a(jz,jgx+5)-drxz_a(jz,jgx-4))+&
!                          b6x*(drxz_a(jz,jgx+6)-drxz_a(jz,jgx-5))
!          end do
!        end do
!!$omp end parallel do
!        paux=0
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., rxz(jz,:),paux(1:nkgx))
!          rxz(jz,:)=paux(1:nkgx)
!        end do
!        paux=0
!        do jz = 1,nkz
!          call interp1d_lop2(.false., .false., pxz(jz,:),paux(1:nkgx))
!          pxz(jz,:)=paux(1:nkgx)
!        end do
!
!        pxz = pxz*costh*sinth
!        rxz = rxz*sinth*costh
!
!!        paux=0
!!        do jgx = 1,nkgx
!!          call interp1d_lop2(.false., .false., pxz(:,jgx),paux(1:nkz))
!!          pxz(:,jgx)=paux(1:nkz)
!!        end do
!!        paux=0
!!        do jgx = 1,nkgx
!!          call interp1d_lop2(.false., .false., rxz(:,jgx),paux(1:nkz))
!!          rxz(:,jgx)=paux(1:nkz)
!!        end do
!        paux=0
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),pxz(:,jgx))
!          pxz(:,jgx)=paux(1:nkz)
!        end do
!        paux=0
!        do jgx = 1,nkgx
!          call interp1d_lop2(.true., .false., paux(1:nkz),rxz(:,jgx))
!          rxz(:,jgx)=paux(1:nkz)
!        end do
!        dpxz_a = pxz
!        drxz_a = rxz
!        pxz=0
!        rxz=0 
!!!$omp parallel do default(shared) &
!!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 6,nkz-6
!            pxz(jz+6,jgx) = pxz(jz+6,jgx)+b6z*dpxz_a(jz,jgx)
!            pxz(jz+5,jgx) = pxz(jz+5,jgx)+b5z*dpxz_a(jz,jgx)
!            pxz(jz+4,jgx) = pxz(jz+4,jgx)+b4z*dpxz_a(jz,jgx)
!            pxz(jz+3,jgx) = pxz(jz+3,jgx)+b3z*dpxz_a(jz,jgx)
!            pxz(jz+2,jgx) = pxz(jz+2,jgx)+b2z*dpxz_a(jz,jgx)
!            pxz(jz+1,jgx) = pxz(jz+1,jgx)+b1z*dpxz_a(jz,jgx)
!            pxz(jz  ,jgx) = pxz(jz  ,jgx)-b1z*dpxz_a(jz,jgx)
!            pxz(jz-1,jgx) = pxz(jz-1,jgx)-b2z*dpxz_a(jz,jgx)
!            pxz(jz-2,jgx) = pxz(jz-2,jgx)-b3z*dpxz_a(jz,jgx)
!            pxz(jz-3,jgx) = pxz(jz-3,jgx)-b4z*dpxz_a(jz,jgx)
!            pxz(jz-4,jgx) = pxz(jz-4,jgx)-b5z*dpxz_a(jz,jgx)
!            pxz(jz-5,jgx) = pxz(jz-5,jgx)-b6z*dpxz_a(jz,jgx)
!
!            rxz(jz+6,jgx) = rxz(jz+6,jgx)+b6z*drxz_a(jz,jgx)
!            rxz(jz+5,jgx) = rxz(jz+5,jgx)+b5z*drxz_a(jz,jgx)
!            rxz(jz+4,jgx) = rxz(jz+4,jgx)+b4z*drxz_a(jz,jgx)
!            rxz(jz+3,jgx) = rxz(jz+3,jgx)+b3z*drxz_a(jz,jgx)
!            rxz(jz+2,jgx) = rxz(jz+2,jgx)+b2z*drxz_a(jz,jgx)
!            rxz(jz+1,jgx) = rxz(jz+1,jgx)+b1z*drxz_a(jz,jgx)
!            rxz(jz  ,jgx) = rxz(jz  ,jgx)-b1z*drxz_a(jz,jgx)
!            rxz(jz-1,jgx) = rxz(jz-1,jgx)-b2z*drxz_a(jz,jgx)
!            rxz(jz-2,jgx) = rxz(jz-2,jgx)-b3z*drxz_a(jz,jgx)
!            rxz(jz-3,jgx) = rxz(jz-3,jgx)-b4z*drxz_a(jz,jgx)
!            rxz(jz-4,jgx) = rxz(jz-4,jgx)-b5z*drxz_a(jz,jgx)
!            rxz(jz-5,jgx) = rxz(jz-5,jgx)-b6z*drxz_a(jz,jgx)
!          end do
!        end do
!!!$omp end parallel do
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz_act(jz,jgx) = dpxz_act(jz,jgx)-1*(pxz(jz,jgx)*fanis1(jz,jgx)&
!                                                  -rxz(jz,jgx)*fanis2(jz,jgx))
!            drxz_act(jz,jgx) = drxz_act(jz,jgx)-1*(pxz(jz,jgx)*fanis2(jz,jgx)&
!                                                  -rxz(jz,jgx))
!            if(ISNAN(dpxz_act(jz,jgx)) .OR. ISNAN(drxz_act(jz,jgx))) then
!              write(0,*) 'NaN in pxz or rxz,jgz,jgx',pxz(jz,jgx),rxz(jz,jgx),jz,jgx
!              stop 666
!            end if
!          end do
!        end do
!      end if
!
!      if(isSrc) then
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshot(jz,jgx)*Vshot(jz,jgx)
!            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshot(jz,jgx)*Vshot(jz,jgx)
!          end do
!        end do
!!$omp end parallel do
!      else
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!        do jgx = 1,nkgx
!          do jz = 1,nkz
!            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshotr(jz,jgx)*Vshotr(jz,jgx)
!            drxz(jz,jgx,jpol) = drxz_act(jz,jgx)*Vshotr(jz,jgx)*Vshotr(jz,jgx)
!          end do
!        end do
!!$omp end parallel do
!      end if
!
!      if(jpol == nB) EXIT
!!$omp parallel do default(shared) &
!!$omp private(jgx,jz) 
!      do jgx = 1,nkgx
!        do jz = 1,nkz
!          dpxz(jz,jgx,jpol+1) = dpxz(jz,jgx,jpol)
!          drxz(jz,jgx,jpol+1) = drxz(jz,jgx,jpol)
!        end do
!      end do
!!$omp end parallel do
!    end do
!  end subroutine

  subroutine lap(ispc,isSrc)
  !*** This subroutine applies the factor -L**2 to     ***
  !*** the wavefield using the rapid expansion method  ***
    implicit none
    integer :: jpol
  
    real :: dp_xx, dr_zz

    logical, intent(in) :: ispc, isSrc

    ppcur => pcur
    ppnew => pnew
    do jpol = 1,nB
      if(ispc) then
        pcur = reshape(dpxz(:,:,jpol),(/nkz*nkgx/))
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
!        do jgx = 7,nkgx-6
!          call props(ppcur,ppnew,nkz,nkgx,jgx-1,dz,dx)
!        end do
!!$omp end parallel do
        dpxz_act(:,:) = reshape(pnew,(/nkz,nkgx/))
      else
        dpxz_cur = dpxz(:,:,jpol)
        dpxz_act = 0
!$omp parallel do default(shared) &
!$omp private(jgx,jz,dp_xx,dr_zz) 
        do jgx = 7,nkgx-6
          do jz = 7,nkz-6
            dp_xx = 0.5*a0*dpxz_cur(jz,jgx)+  &
                    a1x*(dpxz_cur(jz,jgx+1)+dpxz_cur(jz,jgx-1))+&
                    a2x*(dpxz_cur(jz,jgx+2)+dpxz_cur(jz,jgx-2))+&
                    a3x*(dpxz_cur(jz,jgx+3)+dpxz_cur(jz,jgx-3))+&
                    a4x*(dpxz_cur(jz,jgx+4)+dpxz_cur(jz,jgx-4))+&
                    a5x*(dpxz_cur(jz,jgx+5)+dpxz_cur(jz,jgx-5))+&
                    a6x*(dpxz_cur(jz,jgx+6)+dpxz_cur(jz,jgx-6))
      
            dr_zz = 0.5*a0*dpxz_cur(jz,jgx)+  &
                    a1z*(dpxz_cur(jz+1,jgx)+dpxz_cur(jz-1,jgx))+&
                    a2z*(dpxz_cur(jz+2,jgx)+dpxz_cur(jz-2,jgx))+&
                    a3z*(dpxz_cur(jz+3,jgx)+dpxz_cur(jz-3,jgx))+&
                    a4z*(dpxz_cur(jz+4,jgx)+dpxz_cur(jz-4,jgx))+&
                    a5z*(dpxz_cur(jz+5,jgx)+dpxz_cur(jz-5,jgx))+&
                    a6z*(dpxz_cur(jz+6,jgx)+dpxz_cur(jz-6,jgx))
            dpxz_act(jz,jgx) = (dp_xx+dr_zz)
          end do
        end do
!$omp end parallel do
      end if
      if(isSrc) then
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshot(jz,jgx)**2
          end do
        end do
!$omp end parallel do
      else
!$omp parallel do default(shared) &
!$omp private(jgx,jz) 
        do jgx = 1,nkgx
          do jz = 1,nkz
            dpxz(jz,jgx,jpol) = dpxz_act(jz,jgx)*Vshotr(jz,jgx)**2
          end do
        end do
!$omp end parallel do
      end if
      if(jpol == nB) EXIT
      dpxz(:,:,jpol+1) = dpxz(:,:,jpol)
    end do
  end subroutine

  subroutine xcorr(pft,prt,num,den)
    implicit none
    real, dimension(nkz), intent(in) :: pft
    real, dimension(nkz), intent(in) :: prt
    real, dimension(nz), intent(inout) :: num 
    real, dimension(nz), intent(inout) :: den 

    num(:) = num(:)+pft(1+zshift:nz+zshift)*&
                    prt(1+zshift:nz+zshift) 
    den(:) = den(:)+pft(1+zshift:nz+zshift)*&
                    pft(1+zshift:nz+zshift)
!              CONJG(paux_fp(1+zshift:nz+zshift))
  end subroutine

  subroutine clean_all(tti,ispc)
    implicit none
    logical, intent(in) :: tti, ispc 

    deallocate(Vshot, Vshotr)
    if(ispc) then
      if(tti) then
        deallocate(pxxcur)
        deallocate(pzzcur)
        deallocate(pxxnew)
        deallocate(pzznew)
        deallocate(rxxcur)
        deallocate(rzzcur)
        deallocate(rxxnew)
        deallocate(rzznew)
      else
        deallocate(pcur)
        deallocate(pnew)
        deallocate(rcur)
        deallocate(rnew)
      end if
      if(anis) then
        deallocate(drxz_cur)
        if(tti) then
          deallocate(sinth)
          deallocate(costh)
          deallocate(pxz)
          deallocate(rxz)
        end if
      end if
    else
      deallocate(dpxz_cur)
      if(anis) then
        deallocate(drxz_cur)
        if(tti) then
          deallocate(sinth)
          deallocate(costh)
          deallocate(pxz)
          deallocate(rxz)
          deallocate(paux)
          deallocate(dpxz_a)
          deallocate(drxz_a)
        end if
      end if
    end if
  end subroutine

  subroutine fwd_prop_anis(wftm1P,wftm2P,wftP,wftm1R,wftm2R,wftR,ispc,fanis1,fanis2,&
                dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR,tti,energy)
    implicit none
!    optional :: dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR
    integer :: jpol

    real, dimension(:,:) :: fanis1, fanis2
    real, dimension(:,:) :: wftP, wftR, wftm1P, wftm2P, wftm1R, wftm2R
    real, dimension(:,:) :: dpftP, dpftR, dpftPm1, dpftPm2, dpftRm1, dpftRm2
    
    logical, intent(in) :: ispc, tti, energy

    wftm2P(:,:) = wftm1P(:,:)
    wftm2R(:,:) = wftm1R(:,:)
    wftm1P(:,:) = wftP(:,:)
    wftm1R(:,:) = wftR(:,:)
    dpxz = 0
    drxz = 0
    dpxz(:,:,1) = wftm1P
    drxz(:,:,1) = wftm1R
    if(tti) then
      call lap_tti(fanis1,fanis2,ispc,.true.)
    else
      call lap_vti(fanis1,fanis2,ispc,.true.)
    end if
!    if(energy) then
      dpftPm2 = dpftPm1
      dpftPm1 = dpftP
      dpftP(:,:) = dpftPm2(:,:)+2*dt*dpxz(:,:,1)+dt*dt*dt*dpxz(:,:,2)/3
      dpftRm2 = dpftRm1
      dpftRm1 = dpftR
      dpftR(:,:) = dpftRm2(:,:)+2*dt*drxz(:,:,1)+dt*dt*dt*drxz(:,:,2)/3
!    end if
    wftP(:,:) = wftm1P(:,:)*am(0)-wftm2P(:,:)
    wftR(:,:) = wftm1R(:,:)*am(0)-wftm2R(:,:)
    do jpol = 1,nB
      wftP(:,:) = wftP(:,:)+dpxz(:,:,jpol)*am(jpol)
      wftR(:,:) = wftR(:,:)+drxz(:,:,jpol)*am(jpol)
    end do
  end subroutine

  subroutine bwd_propRB_anis(wrtp1P,wrtp2P,wrtP,wrtp1R,wrtp2R,wrtR,ispc,fanis1,fanis2,&
                       dprtPp1,dprtPp2,dprtP,dprtRp1,dprtRp2,dprtR,&
                       wftm1P,wftm2P,wftP,wftm1R,wftm2R,wftR,&
                       dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR,tti,energy)
    implicit none
!    optional :: wftm1P,wftm2P,wftP,wftm1R,wftm2R,wftR
!    optional :: dprtPp1,dprtPp2,dprtP,dprtRp1,dprtRp2,dprtR
!    optional :: dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR

    integer :: jpol

    real, dimension(:,:) :: fanis1, fanis2
    real, dimension(:,:) :: wftP, wftR, wftm1P, wftm2P, wftm1R, wftm2R
    real, dimension(:,:) :: wrtP, wrtR, wrtp1P, wrtp2P, wrtp1R, wrtp2R
    real, dimension(:,:) :: dpftP, dpftR, dpftPm1, dpftPm2, dpftRm1, dpftRm2
    real, dimension(:,:) :: dprtP, dprtR, dprtPp1, dprtPp2, dprtRp1, dprtRp2

    logical, intent(in) :: ispc, tti, energy
!**********************************************
!*** Perform receiver propagation backwards ***
!**********************************************
    wrtp2P(:,:) = wrtp1P(:,:)
    wrtp2R(:,:) = wrtp1R(:,:)
    wrtp1P(:,:) = wrtP(:,:)
    wrtp1R(:,:) = wrtR(:,:)
    dpxz = 0
    drxz = 0
    dpxz(:,:,1) = wrtp1P
    drxz(:,:,1) = wrtp1R
    if(tti) then
      call lap_tti(fanis1,fanis2,ispc,.false.)
    else
      call lap_vti(fanis1,fanis2,ispc,.false.)
    end if
!    if(energy) then
      dprtPp2 = dprtPp1
      dprtPp1 = dprtP
      dprtP(:,:) = dprtPp2(:,:)-2*dt*dpxz(:,:,1)-dt*dt*dt*dpxz(:,:,2)/3
      dprtRp2 = dprtRp1
      dprtRp1 = dprtR
      dprtR(:,:) = dprtRp2(:,:)-2*dt*drxz(:,:,1)-dt*dt*dt*drxz(:,:,2)/3
!    end if
    wrtP(:,:) = wrtp1P(:,:)*am(0)-wrtp2P(:,:)
    wrtR(:,:) = wrtp1R(:,:)*am(0)-wrtp2R(:,:)
    do jpol = 1,nB
      wrtP(:,:) = wrtP(:,:)+dpxz(:,:,jpol)*am(jpol)
      wrtR(:,:) = wrtR(:,:)+drxz(:,:,jpol)*am(jpol)
    end do
!********************************************
!*** Perform source propagation backwards ***
!********************************************
!    if(energy) then
      wftm2P(:,:) = wftm1P(:,:)
      wftm2R(:,:) = wftm1R(:,:)
      wftm1P(:,:) = wftP(:,:)
      wftm1R(:,:) = wftR(:,:)
      dpxz = 0
      drxz = 0
      dpxz(:,:,1) = wftm1P
      drxz(:,:,1) = wftm1R
      if(tti) then
        call lap_tti(fanis1,fanis2,ispc,.true.)
      else
        call lap_vti(fanis1,fanis2,ispc,.true.)
      end if
!      if(energy) then
        dpftPm2 = dpftPm1
        dpftPm1 = dpftP
        dpftP(:,:) = dpftPm2(:,:)-2*dt*dpxz(:,:,1)-dt*dt*dt*dpxz(:,:,2)/3
        dpftRm2 = dpftRm1
        dpftRm1 = dpftR
        dpftR(:,:) = dpftRm2(:,:)-2*dt*drxz(:,:,1)-dt*dt*dt*drxz(:,:,2)/3
!      end if
      wftP(:,:) = wftm1P(:,:)*am(0)-wftm2P(:,:)
      wftR(:,:) = wftm1R(:,:)*am(0)-wftm2R(:,:)
      do jpol = 1,nB
        wftP(:,:) = wftP(:,:)+dpxz(:,:,jpol)*am(jpol)
        wftR(:,:) = wftR(:,:)+drxz(:,:,jpol)*am(jpol)
      end do
!    end if
  end subroutine

  subroutine fwd_prop(wftm1,wftm2,wft,qft,ispc,dpftm1,dpftm2,dpft,energy)
    implicit none
!    optional :: dpftm1,dpftm2,dpft
    integer :: jpol

    real, dimension(:,:) :: qft
    real, dimension(:,:) :: wft, wftm1, wftm2
    real, dimension(:,:) :: dpft, dpftm1, dpftm2
    
    logical, intent(in) :: ispc, energy

    wftm2(:,:) = wftm1(:,:)
    wftm1(:,:) = wft(:,:)
    dpxz = 0
    dpxz(:,:,1) = wftm1+qft
    call lap(ispc,.true.)
!    if(energy) then
      dpftm2 = dpftm1
      dpftm1 = dpft
      dpft(:,:) = dpftm2(:,:)+2*dt*dpxz(:,:,1)+dt*dt*dt*dpxz(:,:,2)/3
!    end if
    wft(:,:) = wftm1(:,:)*am(0)-wftm2(:,:)
    do jpol = 1,nB
      wft(:,:) = wft(:,:)+dpxz(:,:,jpol)*am(jpol)
    end do
  end subroutine

  subroutine bwd_propRB(wrtp1,wrtp2,wrt,qrt,ispc,&
                       dprtp1,dprtp2,dprt,wftm1,wftm2,wft,qft,dpftm1,dpftm2,dpft,energy)
    implicit none
!    optional :: wftm1,wftm2,wft,qft
!    optional :: dprtp1,dprtp2,dprt
!    optional :: dpftm1,dpftm2,dpft

    integer :: jpol

    real, dimension(:,:), intent(in) :: qft, qrt
    real, dimension(:,:), intent(inout) :: wft, wftm1, wftm2
    real, dimension(:,:), intent(inout) :: wrt, wrtp1, wrtp2
    real, dimension(:,:), intent(inout) :: dprt, dprtp1, dprtp2
    real, dimension(:,:), intent(inout) :: dpft, dpftm1, dpftm2

    logical, intent(in) :: ispc, energy
!**********************************************
!*** Perform receiver propagation backwards ***
!**********************************************
    wrtp2(:,:) = wrtp1(:,:)
    wrtp1(:,:) = wrt(:,:)
    dpxz = 0
    dpxz(:,:,1) = wrtp1+qrt
    call lap(ispc,.false.)
!    if(energy) then
      dprtp2 = dprtp1
      dprtp1 = dprt
      dprt(:,:) = dprtp2(:,:)-2*dt*dpxz(:,:,1)-dt*dt*dt*dpxz(:,:,2)/3
!    end if
    wrt(:,:) = wrtp1(:,:)*am(0)-wrtp2(:,:)
    do jpol = 1,nB
      wrt(:,:) = wrt(:,:)+dpxz(:,:,jpol)*am(jpol)
    end do
!********************************************
!*** Perform source propagation backwards ***
!********************************************
!    if(energy) then
      wftm2(:,:) = wftm1(:,:)
      wftm1(:,:) = wft(:,:)
      dpxz = 0
      dpxz(:,:,1) = wftm1+qft
      call lap(ispc,.true.)
!      if(energy) then
        dpftm2 = dpftm1
        dpftm1 = dpft
        dpft(:,:) = dpftm2(:,:)-2*dt*dpxz(:,:,1)-dt*dt*dt*dpxz(:,:,2)/3
!      end if
      wft(:,:) = wftm1(:,:)*am(0)-wftm2(:,:)
      do jpol = 1,nB
        wft(:,:) = wft(:,:)+dpxz(:,:,jpol)*am(jpol)
      end do
!    end if
  end subroutine

  subroutine init_taperCos(pad)
    integer :: jj

    integer, intent(in) :: pad

    real :: dist

    allocate(zxtaper(pad),&
             ctaper(pad,pad))

    do jj = 1,pad
      zxtaper(jj) = 0.99+0.01*COS(pi*(pad-jj)/pad)
    end do

    do jx = 1,pad
      do jz = 1,pad
        dist = SQRT(1.001*(pad-jx)*(pad-jx)/pad/pad+&
                    0.999999*(pad-jz)*(pad-jz)/pad/pad)
        ctaper(jz,jx) = 0.99+0.01*COS(pi*dist) 
      end do
    end do
  end subroutine

  subroutine taperings(prev,cur,next)
    real, dimension(:,:), intent(inout) :: prev,cur,next

    ! Corner
    do jx = 1,pad
      do jz = 1,pad
        prev(jz,jx) =              prev(jz,jx)*ctaper(jz,jx)
        prev(jz,nkgx-jx+1) =       prev(jz,nkgx-jx+1)*ctaper(jz,jx)
        prev(nkz-jz+1,nkgx-jx+1) = prev(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        prev(nkz-jz+1,jx) =        prev(nkz-jz+1,jx)*ctaper(jz,jx)

        cur(jz,jx) =              cur(jz,jx)*ctaper(jz,jx)
        cur(jz,nkgx-jx+1) =       cur(jz,nkgx-jx+1)*ctaper(jz,jx)
        cur(nkz-jz+1,nkgx-jx+1) = cur(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        cur(nkz-jz+1,jx) =        cur(nkz-jz+1,jx)*ctaper(jz,jx)

        next(jz,jx) =              next(jz,jx)*ctaper(jz,jx)
        next(jz,nkgx-jx+1) =       next(jz,nkgx-jx+1)*ctaper(jz,jx)
        next(nkz-jz+1,nkgx-jx+1) = next(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        next(nkz-jz+1,jx) =        next(nkz-jz+1,jx)*ctaper(jz,jx)
      end do
    end do

    do jz = pad+1,nkz-pad
      prev(jz,1:pad) =           prev(jz,1:pad)*zxtaper(1:pad:+1)
      prev(jz,nkgx-pad+1:nkgx) = prev(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)

      cur(jz,1:pad) =           cur(jz,1:pad)*zxtaper(1:pad:+1)
      cur(jz,nkgx-pad+1:nkgx) = cur(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)

      next(jz,1:pad) =           next(jz,1:pad)*zxtaper(1:pad:+1)
      next(jz,nkgx-pad+1:nkgx) = next(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)
    end do

    do jx = pad+1,nkgx-pad
      prev(1:pad,jx) =         prev(1:pad,jx)*zxtaper(1:pad:+1)
      prev(nkz-pad+1:nkz,jx) = prev(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)

      cur(1:pad,jx) =         cur(1:pad,jx)*zxtaper(1:pad:+1)
      cur(nkz-pad+1:nkz,jx) = cur(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)

      next(1:pad,jx) =         next(1:pad,jx)*zxtaper(1:pad:+1)
      next(nkz-pad+1:nkz,jx) = next(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)
    end do
  end subroutine

  subroutine tapering(prev,cur,next)
    real, dimension(:,:), intent(inout) :: prev,cur,next

    ! Corner
    do jx = 1,pad
      do jz = 1,pad
        prev(jz,jx) =              prev(jz,jx)*ctaper(jz,jx)
        prev(jz,nkgx-jx+1) =       prev(jz,nkgx-jx+1)*ctaper(jz,jx)
        prev(nkz-jz+1,nkgx-jx+1) = prev(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        prev(nkz-jz+1,jx) =        prev(nkz-jz+1,jx)*ctaper(jz,jx)

        cur(jz,jx) =              cur(jz,jx)*ctaper(jz,jx)
        cur(jz,nkgx-jx+1) =       cur(jz,nkgx-jx+1)*ctaper(jz,jx)
        cur(nkz-jz+1,nkgx-jx+1) = cur(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        cur(nkz-jz+1,jx) =        cur(nkz-jz+1,jx)*ctaper(jz,jx)

        next(jz,jx) =              next(jz,jx)*ctaper(jz,jx)
        next(jz,nkgx-jx+1) =       next(jz,nkgx-jx+1)*ctaper(jz,jx)
        next(nkz-jz+1,nkgx-jx+1) = next(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
        next(nkz-jz+1,jx) =        next(nkz-jz+1,jx)*ctaper(jz,jx)
      end do
    end do

    do jz = pad+1,nkz-pad
      prev(jz,1:pad) =           prev(jz,1:pad)*zxtaper(1:pad:+1)
      prev(jz,nkgx-pad+1:nkgx) = prev(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)

      cur(jz,1:pad) =           cur(jz,1:pad)*zxtaper(1:pad:+1)
      cur(jz,nkgx-pad+1:nkgx) = cur(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)

      next(jz,1:pad) =           next(jz,1:pad)*zxtaper(1:pad:+1)
      next(jz,nkgx-pad+1:nkgx) = next(jz,nkgx-pad+1:nkgx)*zxtaper(pad:1:-1)
    end do

    do jx = pad+1,nkgx-pad
      prev(1:pad,jx) =         prev(1:pad,jx)*zxtaper(1:pad:+1)
      prev(nkz-pad+1:nkz,jx) = prev(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)

      cur(1:pad,jx) =         cur(1:pad,jx)*zxtaper(1:pad:+1)
      cur(nkz-pad+1:nkz,jx) = cur(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)

      next(1:pad,jx) =         next(1:pad,jx)*zxtaper(1:pad:+1)
      next(nkz-pad+1:nkz,jx) = next(nkz-pad+1:nkz,jx)*zxtaper(pad:1:-1)
    end do
  end subroutine

  function interp1d_lop(adj, add, m, d) result(stat)
    integer             :: stat
    logical, intent(in) :: adj, add
    real, dimension(:)  :: m, d
    call interp1d_lop2(adj, add, m, d)
    stat=0
  end function

  subroutine interp1d_lop2(adj, add, m, d)
  !*** I borrowed this subroutine to interpolate from staggered grid to regular grid
    integer :: jm
    integer :: jd

    real :: fx
    real :: gx

    real, dimension(:) :: m, d

    logical, intent(in) :: adj
    logical, intent(in) :: add

    if(adj) then
      if(.not. add) m=0
    else
      if(.not. add) d=0
    end if
    gx=0.5
    fx=0.5
    if(adj) then
      do jd=1,size(d)
        jm=jd
        if(jm > 1 .and. jm <= size(m)) then
          m(jm-1)= m(jm-1)+gx*d(jd)
          m(jm)  = m(jm)  +fx*d(jd)
        end if
      end do
    else
      do jd=1,size(d)
        jm=jd
        if(jm > 1 .and. jm <= size(m)) then
          d(jd)=d(jd)+gx*m(jm-1)+fx*m(jm)
        end if
      end do
    end if
!      jm=jd-1
!      if(jm >= 1 .and. jm < size(m)) then
!        if(adj) then
!          m(jm)=  m(jm)  +gx*d(jd)
!          m(jm+1)=m(jm+1)+fx*d(jd)
!        else
!          d(jd)=d(jd)+gx*m(jm)+fx*m(jm+1)
!        end if
!      end if
  end subroutine

  function SinC(ang) RESULT(sincf) 
    real :: ang, sincf
    if(ang == 0) then
      sincf = 1.d0 
    else
      sincf = sin(ang)/ang
    end if
  end function
end module
